/* $Id: wberechne.c,v 1.20 1999/01/31 19:35:37 fzk Exp $ */

/* ein Versuch: Nur die Punkte P1,2,3' und P5,6,7' sowie E und I */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "wuerfel.h"

WG *
berechnewg (OPTIONS o, int *anzelt) {
    WG *last = (WG *)NULL;
    WG *x;
    WG *startoflist = (WG *)NULL;

    BWINKEL w;

    if (o.dw == 0.) {
	o.dw = 2. * M_PI / 60.;
    }

    /* w (Winkel) entspricht alpha der Beschreibung. w durchluft die
       Werte 2*PI bis 0 mit einer Schrittweite -dw. Der Startwert 2*PI 
       wird um ein DELTA (= 1/1000) verflscht, um Vielfache von PI zu 
       vermeiden. Fr jede Position wird eine Struktur vom Typ WG
       alloziert, die den Wert w, die Koordinaten der 16 Punkte des
       Wrfelgrtels, sowie die Koordinaten von I und E enthlt. Die
       Indizes haben dabei folgende Bedeutung:

       Index     Bedeutung
       1-8       die obere Hlfte des Wrfelgrtels, Punkte P1 - P8
       9         Evolutionspunkt
       10        Involutionspunkt
       11-18     die unter Hlfte des Wrfelgrtels (Indizes der
                 oberen Hlfte +10). Die Punkte 1,2,5 und 6 sind ident mit 
		 den Punkten 10, 12, 15 und 16
       19        Punkt (2,0,0) auf u-Achse
       20        Punkt (0,2,0) auf v-Achse
       21        Punkt (0,0,2) auf w-Achse
    
       Nach der Berechnung erfolgt die Prfung, ob die Wrfelkanten ==
       1 und die Scharnierkanten == 1/sqrt(3) lange sind (jeweils fr
       eine Kante).  */

    *anzelt = 0;
    for(w = 2. * M_PI - DELTA; w > 0.; w -= o.dw) {
	(*anzelt)++;
	x = (WG *) malloc(sizeof(WG));
	if(startoflist == (WG *)NULL) {
	    startoflist = last = x;
	}
	else {
	    last->nextwgp = x;
	    last = x;
	}
	last->nextwgp =(WG *)NULL;

	berechnewgpos(w, &(last->wgp));
    }

    if (rechnung_nicht_ok) {

	printf ("Achtung: Probe zeigt Rechenfehler, weiter mit RETURN\n");
	warten (10000);
    }
    printwg(o, startoflist);

    return startoflist;
}

void
berechnewgpos(BWINKEL w, WGPOS *wgp) {
    double distanz_p1_e;
    double distanz_p5_i;
    double pdistanz_p1_p3; /* in u/v-Ebene projezierte Strecke p1-p3 */
    double pdistanz_p5_p7; /* in u/v-Ebene projezierte Strecke p5-p7 */
    double eu;
    double ix;
    double s;
    int i;
    double laenge, faktor;
    BWINKEL e1, i1, e5, i5, e51;

    wgp->w = w;

    eu = 2. / sqrt(3.);    /* u-Koordinate des Evolutionspunktes */
    ix = -2. / sqrt(3.);
    s = 1. / sqrt(3.);     /* Lnge einer Scharnierkante */

    /* Die 4 Winkel e1, i1, e5, i5 berechnen.
    ** e1 ist der Winkel zwischen u-Achse und der Strecke p1-e
    ** i1                         u-Achse                 p1-i
    ** e5                         u-Achse                 p5-e
    ** i5                         u-Achse                 p5-i
    */
    if (w > (3./2. *M_PI)) {
	e1 = w - 3./2.*M_PI;
    }
    else if (w > M_PI) {
	e1 = w - 3./2.*M_PI;
    }
    else if (w > M_PI_2) {
	e1 = w - M_PI_2;
    }
    else {
	e1 = w - M_PI_2;
    }
    i1 = w;

    /* e5 liegt vorerst nur im Bereich -PI/2 - +PI/2 und mu noch korrigiert
    ** werden
    */
    e51 = atan (eu * tan(e1) / 1. / sqrt(3.));

    if (w > (3./2.*M_PI)) {
	e5 = M_PI - e51;
	i5 = e5 + M_PI_2;
    }
    else if (w > M_PI) {
	e5 =M_PI - e51;
	i5 = e5 - M_PI_2;
    }
    else if (w > M_PI_2) {
	e5 = 2.*M_PI - e51;
	i5 = e5 + M_PI_2;
    }
    else {
	e5 = 2.*M_PI - e51;
	i5 = e5 - M_PI_2;
    }

/*    fprintf (fw, "xx w=%4.0f e1=%4.0f i1=%4.0f i5=%4.0f e5=%4.0f e51=%4.0f\n", */
/* 	       GWINKEL(w), GWINKEL(e1), GWINKEL(i1), GWINKEL(i5), GWINKEL(e5), */
/* 	       GWINKEL(e51)); */

    /*
    ** linker Teil des Wuerfelguertels
    */
    KP(1,u) = 0;
    KP(1,v) = 0;
    KP(1,w) = 0;

    KP(2,u) = s * cos(i1);
    KP(2,v) = s * sin(i1);
    KP(2,w) = 0;

    /*
    ** Evolutionspunkt
    */
    KP(9,u) = eu;
    KP(9,v) = eu * tan(e1);
    KP(9,w) = 0;

    distanz_p1_e = eu / cos(e1);
    pdistanz_p1_p3 = 1. / distanz_p1_e;

    KP(3,u) = pdistanz_p1_p3 * cos(e1);
    KP(3,v) = pdistanz_p1_p3 * sin(e1);
    KP(3,w) = sqrt(1. - pow(pdistanz_p1_p3,2.));

    KP(4,u) = KP(2,v)*KP(3,w) - KP(2,w)*KP(3,v);
    KP(4,v) = KP(2,w)*KP(3,u) - KP(2,u)*KP(3,w);
    KP(4,w) = KP(2,u)*KP(3,v) - KP(2,v)*KP(3,u);

    KP(4,u) = KP(3,u) - KP(4,u);
    KP(4,v) = KP(3,v) - KP(4,v);
    KP(4,w) = KP(3,w) - KP(4,w);

    /*
    ** Involutionspunkt
    */
    KP(10,u) = s;
    KP(10,v) = s * tan (i1);
    KP(10,w) = 0.;

    /*   distanz_p5_i = ix / cos(i5); */
    distanz_p5_i = sqrt (pow (KP(10,v), 2.) + pow (2. / sqrt (3.), 2.));
    if (w < M_PI) {
	distanz_p5_i = -distanz_p5_i;
    }
    /* e5 = i5 + M_PI_2; */
 
    /*
    ** rechter Teil des Wuerfelguertels
    ** P5 wird als im 0-Punkt liegend betrachtet
    */
    KP(5,u) = sqrt(3.);
    KP(5,v) = 0.;
    KP(5,w) = 0.;

    /* Die Berechnung von P6 erfolgt hier abweichend von der Beschreibung.
    ** P6 wird so ermittelt, da auf der Geraden P5-E eine Strecke der Lnge
    ** 1/sqrt(3) berechnet wird.
    */
    KP(6,u) = KP(5,u) - KP(9,u);
    KP(6,v) = KP(5,v) - KP(9,v);
    KP(6,w) = 0.;
    laenge = sqrt(pow(KP(6,u),2.) + pow(KP(6,v),2.) + pow(KP(6,w),2.));
    faktor = 1. / (sqrt(3.) * laenge);
    if (w > M_PI) {faktor = -faktor;}
    KP(6,u) = KP(6,u) * faktor;
    KP(6,v) = KP(6,v) * faktor;
    KP(6,w) = 0.;

    /* Die Berechnung von P7 erfolgt hier abweichend von der Beschreibung.
    ** P7 wird so ermittelt, da auf der Geraden P5-I eine Strecke der Lnge
    ** projezierte Distanz P5-P7 berechnet wird.
    */

    pdistanz_p5_p7 = 1. / distanz_p5_i;

    KP(7,u) = KP(5,u) - KP(10,u);
    KP(7,v) = KP(5,v) - KP(10,v);
    KP(7,w) = KP(5,w) - KP(10,w);
    laenge = sqrt(pow(KP(7,u),2.) + pow(KP(7,v),2.) + pow(KP(7,w),2.));
    faktor = pdistanz_p5_p7 / laenge;
    if (w > M_PI) {faktor = -faktor;}
    KP(7,u) = KP(7,u) * faktor;
    KP(7,v) = KP(7,v) * faktor;
    KP(7,w) = sqrt(1. - pow(pdistanz_p5_p7,2.));


    KP(8,u) = KP(6,v)*KP(7,w) - KP(6,w)*KP(7,v);
    KP(8,v) = KP(6,w)*KP(7,u) - KP(6,u)*KP(7,w);
    KP(8,w) = KP(6,u)*KP(7,v) - KP(6,v)*KP(7,u);

    KP(8,u) = KP(7,u) - KP(8,u);
    KP(8,v) = KP(7,v) - KP(8,v);
    KP(8,w) = KP(7,w) - KP(8,w);

    /*
    **  den rechten Teil des Wuerfelguertels um sqrt(3.) nach rechts schieben
    */
    /*   KP(5,u) = KP(5,u) + sqrt(3.); */
    KP(6,u) = KP(6,u) + sqrt(3.);
    KP(7,u) = KP(7,u) + sqrt(3.);
    KP(8,u) = KP(8,u) + sqrt(3.);

    /*
    **  unterer Teil des Wuerfelguertels. Spiegelung um die u/v-Ebene, also
    **  w-Koordinate negativ.
    */
    for (i=1; i<=8; i++) {
	KP(i+10,u) =  KP(i,u);
	KP(i+10,v) =  KP(i,v);
	KP(i+10,w) = -KP(i,w);
    }

    /* Auf jeder Koordinatenachse ein Punkt markieren. Die Strecken P1-P19,
    ** P1-P20 und P1-P21 bilden die Koordinatenachsen
    */
    KP(19,u) = 2.;
    KP(19,v) = 0.;
    KP(19,w) = 0.;
    KP(20,u) = 0.;
    KP(20,v) = 2.;
    KP(20,w) = 0.;
    KP(21,u) = 0.;
    KP(21,v) = 0.;
    KP(21,w) = 2.;

    /*
    ** Probe 1: Distanz P3-P7 mu 1 sein
    */
    if ((1. - sqrt (pow (KP(3,u)-KP(7,u), 2.) +
                    pow (KP(3,v)-KP(7,v), 2.) +
                    pow (KP(3,w)-KP(7,w), 2.))) > DELTA) {
	printf ("Distanz P3-P7 bei %7.3f ist %f soll %f\n",
		GWINKEL(w),
		sqrt (pow (KP(3,u)-KP(7,u), 2.) +
		      pow (KP(3,v)-KP(7,v), 2.) +
		      pow (KP(3,w)-KP(7,w), 2.)), 1.);
	rechnung_nicht_ok = TRUE;
    }
    /*
    ** Probe 2: Distanz P7-P8 mu 1/sqrt(3) sein
    */
    if ((s - sqrt (pow (KP(6,u)-KP(5,u), 2.) +
		   pow (KP(6,v)-KP(5,v), 2.) +
		   pow (KP(6,w)-KP(5,w), 2.))) > DELTA) {
	printf ("Distanz P5-P6 bei %7.3f ist %f soll %f\n",
		GWINKEL(w),
		sqrt (pow (KP(6,u)-KP(5,u), 2.) +
		      pow (KP(6,v)-KP(5,v), 2.) +
		      pow (KP(6,w)-KP(5,w), 2.)), s);
	rechnung_nicht_ok = TRUE;
    }
}

#define PRINT2D(i1, i2, f,n) (fprintf (f##i1##i2, "%f %f\n",  \
				       twgp->wgp.e[n-1].i1, \
				       twgp->wgp.e[n-1].i2))
#define PRINT3D(f,n) (fprintf (f, "%f %f %f\n", twgp->wgp.e[n-1].u, \
                                                twgp->wgp.e[n-1].v, \
						twgp->wgp.e[n-1].w))
#define PRINTSEGS(f,k) (fprintf (f, "%f %f %f\n%f %f %f\n", \
        twgp->wgp.e[(kante[k-1].von)-1].u, \
        twgp->wgp.e[(kante[k-1].von)-1].v, \
        twgp->wgp.e[(kante[k-1].von)-1].w, \
        twgp->wgp.e[(kante[k-1].nach)-1].u, \
	twgp->wgp.e[(kante[k-1].nach)-1].v, \
	twgp->wgp.e[(kante[k-1].nach)-1].w))

void
printwg(OPTIONS o, WG *wg) {
    WG *twgp;
    int i, j;
    FILE *f2duv;
    FILE *f2duw;
    FILE *f2dvw;
    FILE *f3d;
    FILE *fsegs;

    if (! o.print) {
	return;
    }

    f2duv = fopen (FILE2D(u, v), "w"); /* Files werden ueberschrieben,
					      wenn sie bereits existieren */
    f2duw = fopen (FILE2D(u, w), "w");
    f2dvw = fopen (FILE2D(v, w), "w");
    f3d   = fopen (FILE3D, "w");
    fsegs = fopen (FILESEGS, "w");

    for(twgp=wg; twgp->nextwgp !=(WG *)NULL; twgp=twgp->nextwgp) {
	fprintf (f2duv, "# Position bei Winkel %7.3f\n", GWINKEL(twgp->wgp.w));
	fprintf (f2duw, "# Position bei Winkel %7.3f\n", GWINKEL(twgp->wgp.w));
	fprintf (f2dvw, "# Position bei Winkel %7.3f\n", GWINKEL(twgp->wgp.w));
	fprintf (f3d,   "# Position bei Winkel %7.3f\n", GWINKEL(twgp->wgp.w));
	fprintf (fsegs, "# Position bei Winkel %7.3f\n", GWINKEL(twgp->wgp.w));

	/*
	** nur die obere Hlfte des WG wird ausgegeben
	*/
	for (i=1; i<=8; i++) {
	    PRINT2D (u, v, f2d, i);
	    PRINT2D (u, w, f2d, i);
	    PRINT2D (v, w, f2d, i);
	    PRINT3D (f3d, i);
	}

	/*
	** alle in wuerfel.h definierten Kanten werden ausgegeben
	*/
	for (j=1; j<=anzkanten; j++) {
	    PRINTSEGS (fsegs,j);
	}
	
	fprintf (f2duv, "\n");
	fprintf (f2duw, "\n");
	fprintf (f2dvw, "\n");
	fprintf (f3d,   "\n");
	fprintf (fsegs, "\n\n");
    }
    fclose (f2duv);
    fclose (f2duw);
    fclose (f2dvw);
    fclose (f3d);
    fclose (fsegs);
}

void
freewg (WG *wg) {
    WG *wgp;

    wgp = wg;
    while (wgp) {
	WG *t;

	t = wgp;
	wgp = wgp->nextwgp;
	free (t);
    }
}
